"""
This file is auto generated. Do not edit this file manually.
Source code: [monorepo]/contracts/generator.py
"""

import typing
from dataclasses import dataclass

CloudApiCliApiKeyInfoResponseAuthorId = str

CloudApiCliApiKeyInfoResponseProjectId = str

CloudApiCliApiKeyInfoResponseEmail = str

CloudApiCliApiKeyInfoResponseIntercomHash = str


@dataclass
class CloudApiCliApiKeyInfoResponse:
    author_id: CloudApiCliApiKeyInfoResponseAuthorId
    project_id: CloudApiCliApiKeyInfoResponseProjectId
    email: CloudApiCliApiKeyInfoResponseEmail
    intercom_hash: CloudApiCliApiKeyInfoResponseIntercomHash

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "authorId": self.author_id,
            "projectId": self.project_id,
            "email": self.email,
            "intercomHash": self.intercom_hash,
        }

    @staticmethod
    def from_dict(
        data: typing.Dict[str, typing.Any],
    ) -> "CloudApiCliApiKeyInfoResponse":
        return CloudApiCliApiKeyInfoResponse(
            author_id=str(data["authorId"]),
            project_id=str(data["projectId"]),
            email=str(data["email"]),
            intercom_hash=str(data["intercomHash"]),
        )


CommonRoleDescription = str

CloudApiCliRoleUpdateRequestDescription = CommonRoleDescription


@dataclass
class CloudApiCliRoleUpdateRequest:
    description: CloudApiCliRoleUpdateRequestDescription

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "description": self.description,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "CloudApiCliRoleUpdateRequest":
        return CloudApiCliRoleUpdateRequest(
            description=str(data["description"]),
        )


CommonRoleId = str

CommonRoleName = str

CommonRoleProjectId = str

CommonRoleCreatedAt = str


@dataclass
class CommonRole:
    id: CommonRoleId
    name: CommonRoleName
    description: typing.Optional[CommonRoleDescription]
    project_id: CommonRoleProjectId
    created_at: CommonRoleCreatedAt

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "projectId": self.project_id,
            "createdAt": self.created_at,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "CommonRole":
        return CommonRole(
            id=str(data["id"]),
            name=str(data["name"]),
            description=str(data.get("description")),
            project_id=str(data["projectId"]),
            created_at=str(data["createdAt"]),
        )


CloudApiCliRoleUpdateResponse = CommonRole

CloudApiCliRoleListQueryOffset = int

CloudApiCliRoleListQueryLimit = int


@dataclass
class CloudApiCliRoleListQuery:
    offset: typing.Optional[CloudApiCliRoleListQueryOffset]
    limit: typing.Optional[CloudApiCliRoleListQueryLimit]

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "offset": self.offset,
            "limit": self.limit,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "CloudApiCliRoleListQuery":
        return CloudApiCliRoleListQuery(
            offset=int(data.get("offset", 0)),
            limit=int(data.get("limit", 10)),
        )


CloudApiCliRoleListResponseItem = CommonRole

CloudApiCliRoleListResponse = typing.List[CloudApiCliRoleListResponseItem]

CloudApiCliRoleCreateRequestName = CommonRoleName

CloudApiCliRoleCreateRequestDescription = CommonRoleDescription


@dataclass
class CloudApiCliRoleCreateRequest:
    name: CloudApiCliRoleCreateRequestName
    description: typing.Optional[CloudApiCliRoleCreateRequestDescription]

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "name": self.name,
            "description": self.description,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "CloudApiCliRoleCreateRequest":
        return CloudApiCliRoleCreateRequest(
            name=str(data["name"]),
            description=str(data.get("description")),
        )


CloudApiCliRoleCreateResponse = CommonRole

CommonUserId = str

CommonUserEmail = str

CommonUserRolesItem = str

CommonUserRoles = typing.List[CommonUserRolesItem]

CommonUserProjectId = str

CommonUserCreatedAt = str


@dataclass
class CommonUser:
    id: CommonUserId
    email: CommonUserEmail
    roles: CommonUserRoles
    project_id: CommonUserProjectId
    created_at: CommonUserCreatedAt

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "id": self.id,
            "email": self.email,
            "roles": self.roles,
            "projectId": self.project_id,
            "createdAt": self.created_at,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "CommonUser":
        return CommonUser(
            id=str(data["id"]),
            email=str(data["email"]),
            roles=[str(item) for item in data["roles"]],
            project_id=str(data["projectId"]),
            created_at=str(data["createdAt"]),
        )


CloudApiCliUserGetResponse = CommonUser

CloudApiCliBuildCreateResponseUrl = str

CloudApiCliBuildCreateResponseBuildId = str


@dataclass
class CloudApiCliBuildCreateResponse:
    url: CloudApiCliBuildCreateResponseUrl
    build_id: CloudApiCliBuildCreateResponseBuildId

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "url": self.url,
            "buildId": self.build_id,
        }

    @staticmethod
    def from_dict(
        data: typing.Dict[str, typing.Any],
    ) -> "CloudApiCliBuildCreateResponse":
        return CloudApiCliBuildCreateResponse(
            url=str(data["url"]),
            build_id=str(data["buildId"]),
        )


AbstraLibApiStageCardContentKey = str

AbstraLibApiStageCardContentValue = typing.Dict[str, typing.Any]

AbstraLibApiStageCardContentType = str


@dataclass
class AbstraLibApiStageCardContent:
    key: AbstraLibApiStageCardContentKey
    value: AbstraLibApiStageCardContentValue
    type: AbstraLibApiStageCardContentType

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "key": self.key,
            "value": self.value,
            "type": self.type,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "AbstraLibApiStageCardContent":
        return AbstraLibApiStageCardContent(
            key=str(data["key"]),
            value=dict(**data["value"]),
            type=str(data["type"]),
        )


AbstraLibApiStageCardContentsItem = AbstraLibApiStageCardContent

AbstraLibApiStageCardContents = typing.List[AbstraLibApiStageCardContentsItem]

AbstraLibApiStageRunId = str

AbstraLibApiStageRunStage = str

AbstraLibApiStageRunData = typing.Dict[str, typing.Any]

AbstraLibApiStageRunStatus = typing.Union[
    typing.Literal["waiting"],
    typing.Literal["running"],
    typing.Literal["processing"],
    typing.Literal["finished"],
    typing.Literal["failed"],
    typing.Literal["abandoned"],
]

AbstraLibApiStageRunStatusValues: typing.List[AbstraLibApiStageRunStatus] = [
    "waiting",
    "running",
    "processing",
    "finished",
    "failed",
    "abandoned",
]

AbstraLibApiStageRunCreatedAt = str

AbstraLibApiStageRunUpdatedAt = str

AbstraLibApiStageRunParentId = str

AbstraLibApiStageRunExecutionId = str


@dataclass
class AbstraLibApiStageRun:
    id: AbstraLibApiStageRunId
    stage: AbstraLibApiStageRunStage
    data: AbstraLibApiStageRunData
    status: AbstraLibApiStageRunStatus
    created_at: AbstraLibApiStageRunCreatedAt
    updated_at: AbstraLibApiStageRunUpdatedAt
    parent_id: typing.Optional[AbstraLibApiStageRunParentId]
    execution_id: typing.Optional[AbstraLibApiStageRunExecutionId]

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "id": self.id,
            "stage": self.stage,
            "data": self.data,
            "status": self.status,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
            "parent_id": self.parent_id,
            "execution_id": self.execution_id,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "AbstraLibApiStageRun":
        return AbstraLibApiStageRun(
            id=str(data["id"]),
            stage=str(data["stage"]),
            data=dict(**data["data"]),
            status=data["status"],
            created_at=str(data["created_at"]),
            updated_at=str(data["updated_at"]),
            parent_id=str(data.get("parent_id")),
            execution_id=str(data.get("execution_id")),
        )


AbstraLibApiStageRunCardId = str

AbstraLibApiStageRunCardCreatedAt = str

AbstraLibApiStageRunCardUpdatedAt = str

AbstraLibApiStageRunCardAssignee = str

AbstraLibApiStageRunCardStatus = AbstraLibApiStageRunStatus

AbstraLibApiStageRunCardContent = AbstraLibApiStageCardContents

AbstraLibApiStageRunCardStage = str


@dataclass
class AbstraLibApiStageRunCard:
    id: AbstraLibApiStageRunCardId
    created_at: AbstraLibApiStageRunCardCreatedAt
    updated_at: typing.Optional[AbstraLibApiStageRunCardUpdatedAt]
    assignee: typing.Optional[AbstraLibApiStageRunCardAssignee]
    status: AbstraLibApiStageRunCardStatus
    content: AbstraLibApiStageRunCardContent
    stage: typing.Optional[AbstraLibApiStageRunCardStage]

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "id": self.id,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
            "assignee": self.assignee,
            "status": self.status,
            "content": self.content,
            "stage": self.stage,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "AbstraLibApiStageRunCard":
        return AbstraLibApiStageRunCard(
            id=str(data["id"]),
            created_at=str(data["created_at"]),
            updated_at=str(data.get("updated_at")),
            assignee=str(data.get("assignee")),
            status=data["status"],
            content=[
                AbstraLibApiStageCardContent.from_dict(item) for item in data["content"]
            ],
            stage=str(data.get("stage")),
        )


AbstraLibApiStageRunCardsItem = AbstraLibApiStageRunCard

AbstraLibApiStageRunCards = typing.List[AbstraLibApiStageRunCardsItem]

AbstraLibApiStageId = str

AbstraLibApiStageType = typing.Union[
    typing.Literal["form"],
    typing.Literal["hook"],
    typing.Literal["job"],
    typing.Literal["script"],
    typing.Literal["iterator"],
    typing.Literal["condition"],
]

AbstraLibApiStageTypeValues: typing.List[AbstraLibApiStageType] = [
    "form",
    "hook",
    "job",
    "script",
    "iterator",
    "condition",
]

AbstraLibApiStageTitle = str

AbstraLibApiStagePath = str

AbstraLibApiStageCanBeStarted = bool


@dataclass
class AbstraLibApiStage:
    id: AbstraLibApiStageId
    type: AbstraLibApiStageType
    title: AbstraLibApiStageTitle
    path: typing.Optional[AbstraLibApiStagePath]
    can_be_started: AbstraLibApiStageCanBeStarted

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "id": self.id,
            "type": self.type,
            "title": self.title,
            "path": self.path,
            "can_be_started": self.can_be_started,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "AbstraLibApiStage":
        return AbstraLibApiStage(
            id=str(data["id"]),
            type=data["type"],
            title=str(data["title"]),
            path=str(data.get("path")),
            can_be_started=bool(data["can_be_started"]),
        )


AbstraLibApiKanbanColumnSelectedStage = AbstraLibApiStage

AbstraLibApiKanbanColumnStageRunCards = AbstraLibApiStageRunCards

AbstraLibApiKanbanColumnTotalCount = int

AbstraLibApiKanbanColumnLoading = bool


@dataclass
class AbstraLibApiKanbanColumn:
    selected_stage: AbstraLibApiKanbanColumnSelectedStage
    stage_run_cards: AbstraLibApiKanbanColumnStageRunCards
    total_count: AbstraLibApiKanbanColumnTotalCount
    loading: typing.Optional[AbstraLibApiKanbanColumnLoading]

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "selected_stage": self.selected_stage,
            "stage_run_cards": self.stage_run_cards,
            "total_count": self.total_count,
            "loading": self.loading,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "AbstraLibApiKanbanColumn":
        return AbstraLibApiKanbanColumn(
            selected_stage=AbstraLibApiStage.from_dict(data["selected_stage"]),
            stage_run_cards=[
                AbstraLibApiStageRunCard.from_dict(item)
                for item in data["stage_run_cards"]
            ],
            total_count=int(data["total_count"]),
            loading=bool(data.get("loading")),
        )


AbstraLibApiKanbanColumnsItem = AbstraLibApiKanbanColumn

AbstraLibApiKanbanColumns = typing.List[AbstraLibApiKanbanColumnsItem]

AbstraLibApiKanbanDataColumns = AbstraLibApiKanbanColumns


@dataclass
class AbstraLibApiKanbanData:
    columns: AbstraLibApiKanbanDataColumns

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "columns": self.columns,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "AbstraLibApiKanbanData":
        return AbstraLibApiKanbanData(
            columns=[
                AbstraLibApiKanbanColumn.from_dict(item) for item in data["columns"]
            ],
        )


AbstraLibApiThreadsStageRunCards = AbstraLibApiStageRunCards

AbstraLibApiThreadsNotFoundStagesItem = str

AbstraLibApiThreadsNotFoundStages = typing.List[AbstraLibApiThreadsNotFoundStagesItem]

AbstraLibApiThreadsTotalCount = int


@dataclass
class AbstraLibApiThreads:
    stage_run_cards: AbstraLibApiThreadsStageRunCards
    not_found_stages: AbstraLibApiThreadsNotFoundStages
    total_count: AbstraLibApiThreadsTotalCount

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "stage_run_cards": self.stage_run_cards,
            "not_found_stages": self.not_found_stages,
            "total_count": self.total_count,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "AbstraLibApiThreads":
        return AbstraLibApiThreads(
            stage_run_cards=[
                AbstraLibApiStageRunCard.from_dict(item)
                for item in data["stage_run_cards"]
            ],
            not_found_stages=[str(item) for item in data["not_found_stages"]],
            total_count=int(data["total_count"]),
        )


AbstraLibApiStagesItem = AbstraLibApiStage

AbstraLibApiStages = typing.List[AbstraLibApiStagesItem]

AbstraLibApiEditorLintersRuleName = str

AbstraLibApiEditorLintersRuleLabel = str

AbstraLibApiEditorLintersRuleType = typing.Literal["bug"]

AbstraLibApiEditorLintersRuleTypeValues: typing.List[
    AbstraLibApiEditorLintersRuleType
] = ["bug"]

AbstraLibApiEditorLintersIssueLabel = str

AbstraLibApiEditorLintersFixName = str

AbstraLibApiEditorLintersFixLabel = str


@dataclass
class AbstraLibApiEditorLintersFix:
    name: AbstraLibApiEditorLintersFixName
    label: AbstraLibApiEditorLintersFixLabel

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "name": self.name,
            "label": self.label,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "AbstraLibApiEditorLintersFix":
        return AbstraLibApiEditorLintersFix(
            name=str(data["name"]),
            label=str(data["label"]),
        )


AbstraLibApiEditorLintersIssueFixesItem = AbstraLibApiEditorLintersFix

AbstraLibApiEditorLintersIssueFixes = typing.List[
    AbstraLibApiEditorLintersIssueFixesItem
]


@dataclass
class AbstraLibApiEditorLintersIssue:
    label: AbstraLibApiEditorLintersIssueLabel
    fixes: AbstraLibApiEditorLintersIssueFixes

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "label": self.label,
            "fixes": self.fixes,
        }

    @staticmethod
    def from_dict(
        data: typing.Dict[str, typing.Any],
    ) -> "AbstraLibApiEditorLintersIssue":
        return AbstraLibApiEditorLintersIssue(
            label=str(data["label"]),
            fixes=[
                AbstraLibApiEditorLintersFix.from_dict(item) for item in data["fixes"]
            ],
        )


AbstraLibApiEditorLintersRuleIssuesItem = AbstraLibApiEditorLintersIssue

AbstraLibApiEditorLintersRuleIssues = typing.List[
    AbstraLibApiEditorLintersRuleIssuesItem
]


@dataclass
class AbstraLibApiEditorLintersRule:
    name: AbstraLibApiEditorLintersRuleName
    label: AbstraLibApiEditorLintersRuleLabel
    type: AbstraLibApiEditorLintersRuleType
    issues: AbstraLibApiEditorLintersRuleIssues

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "name": self.name,
            "label": self.label,
            "type": self.type,
            "issues": self.issues,
        }

    @staticmethod
    def from_dict(
        data: typing.Dict[str, typing.Any],
    ) -> "AbstraLibApiEditorLintersRule":
        return AbstraLibApiEditorLintersRule(
            name=str(data["name"]),
            label=str(data["label"]),
            type=data["type"],
            issues=[
                AbstraLibApiEditorLintersIssue.from_dict(item)
                for item in data["issues"]
            ],
        )


AbstraLibApiEditorEnvVarsModelName = str

AbstraLibApiEditorEnvVarsModelValue = str


@dataclass
class AbstraLibApiEditorEnvVarsModel:
    name: AbstraLibApiEditorEnvVarsModelName
    value: AbstraLibApiEditorEnvVarsModelValue

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "name": self.name,
            "value": self.value,
        }

    @staticmethod
    def from_dict(
        data: typing.Dict[str, typing.Any],
    ) -> "AbstraLibApiEditorEnvVarsModel":
        return AbstraLibApiEditorEnvVarsModel(
            name=str(data["name"]),
            value=str(data["value"]),
        )


AbstraLibApiEditorLintersFixResponseSuccess = bool


@dataclass
class AbstraLibApiEditorLintersFixResponse:
    success: AbstraLibApiEditorLintersFixResponseSuccess

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "success": self.success,
        }

    @staticmethod
    def from_dict(
        data: typing.Dict[str, typing.Any],
    ) -> "AbstraLibApiEditorLintersFixResponse":
        return AbstraLibApiEditorLintersFixResponse(
            success=bool(data["success"]),
        )


AbstraLibApiEditorLintersCheckResponseItem = AbstraLibApiEditorLintersRule

AbstraLibApiEditorLintersCheckResponse = typing.List[
    AbstraLibApiEditorLintersCheckResponseItem
]

AbstraLibApiEditorEnvVarsListResponseItem = AbstraLibApiEditorEnvVarsModel

AbstraLibApiEditorEnvVarsListResponse = typing.List[
    AbstraLibApiEditorEnvVarsListResponseItem
]

AbstraLibApiEditorEnvVarsCreateRequest = AbstraLibApiEditorEnvVarsModel

AbstraLibApiEditorEnvVarsCreateResponse = AbstraLibApiEditorEnvVarsModel

AbstraLibApiPlayerUserSignupEmail = CommonUserEmail


@dataclass
class AbstraLibApiPlayerUserSignup:
    email: AbstraLibApiPlayerUserSignupEmail

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "email": self.email,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "AbstraLibApiPlayerUserSignup":
        return AbstraLibApiPlayerUserSignup(
            email=str(data["email"]),
        )


AbstraLibApiPlayerUserNavigationGuardStatus = typing.Union[
    typing.Literal["ALLOWED"],
    typing.Literal["UNAUTHORIZED"],
    typing.Literal["FORBIDEN"],
    typing.Literal["NOT_FOUND"],
]

AbstraLibApiPlayerUserNavigationGuardStatusValues: typing.List[
    AbstraLibApiPlayerUserNavigationGuardStatus
] = ["ALLOWED", "UNAUTHORIZED", "FORBIDEN", "NOT_FOUND"]


@dataclass
class AbstraLibApiPlayerUserNavigationGuard:
    status: AbstraLibApiPlayerUserNavigationGuardStatus

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "status": self.status,
        }

    @staticmethod
    def from_dict(
        data: typing.Dict[str, typing.Any],
    ) -> "AbstraLibApiPlayerUserNavigationGuard":
        return AbstraLibApiPlayerUserNavigationGuard(
            status=data["status"],
        )



import sys

from abstra_internals.controllers.execution_stdio import StdioController
from abstra_internals.controllers.main import MainController
from abstra_internals.environment import DISABLE_STDIO_PATCH


class UnboundPthread(Exception):
    pass


class StdioPatcher:
    original_sys_stdout_write = sys.stdout.write
    original_sys_stderr_write = sys.stderr.write

    @classmethod
    def apply(cls, main_controller: MainController):
        if DISABLE_STDIO_PATCH:
            return

        stdio_controller = StdioController(
            sys_stderr_write=cls.original_sys_stderr_write,
            sys_stdout_write=cls.original_sys_stdout_write,
            main_controller=main_controller,
        )

        sys.stdout.write = stdio_controller.patched_stdout_write
        sys.stderr.write = stdio_controller.patched_stderr_write

    @classmethod
    def reset(cls):
        sys.stdout.write = cls.original_sys_stdout_write
        sys.stderr.write = cls.original_sys_stderr_write



import unittest

from abstra_internals.linter.rules.conflicting_path import ConflictingPath
from abstra_internals.repositories.project.project import FormStage, ProjectRepository
from tests.fixtures import clear_dir, init_dir


class ConflictingPathTest(unittest.TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_conflicting_path_valid_default(self):
        rule = ConflictingPath()
        self.assertEqual(len(rule.find_issues()), 0)

    def test_conflicting_path_without_conflict(self):
        project = ProjectRepository.load()
        form = FormStage.create(
            id="test",
            title="test",
            file="test.py",
        )
        form.path = "not_conflicting_path"
        project.add_stage(form)
        ProjectRepository.save(project)

        rule = ConflictingPath()
        self.assertEqual(len(rule.find_issues()), 0)

    def test_conflicting_path_with_conflict(self):
        project = ProjectRepository.load()
        form = FormStage.create(
            id="login",
            title="login",
            file="login.py",
        )
        form.path = "login"
        project.add_stage(form)
        ProjectRepository.save(project)

        rule = ConflictingPath()
        issues = rule.find_issues()
        self.assertEqual(len(issues), 1)

        issue = issues[0]
        self.assertEqual(len(issue.fixes), 1)

        fix = issue.fixes[0]
        fix.fix()
        self.assertEqual(len(rule.find_issues()), 0)



import io
import pathlib
import shutil
import typing as t

from werkzeug.datastructures import FileStorage

from abstra_internals.utils.file import get_random_filepath


def public_path(name: str):
    return "/_files/" + name


def upload_file(file: t.Union[FileStorage, io.IOBase, pathlib.Path]):
    if isinstance(file, (io.BufferedReader, io.TextIOWrapper)):
        # TODO: use hasattr(file, "name")
        name, path = get_random_filepath(file.name)
        shutil.copy(file.name, path)
        return public_path(name)

    if isinstance(file, FileStorage):
        name, path = get_random_filepath(file.name)
        file.save(path)
        return public_path(name)

    if isinstance(file, pathlib.Path):
        name, path = get_random_filepath(file.name)
        shutil.copy(file, path)
        return public_path(name)

    if isinstance(file, io.IOBase):
        name, path = get_random_filepath()
        with open(path, "wb") as f:
            shutil.copyfileobj(file, f)
        return public_path(name)

    raise ValueError(f"Cannot upload {type(file)}")



from abstra_internals.widgets.widget_base import Output


class TextOutput(Output):
    type = 'text-output'

    def __init__(self, text: str, **kwargs):
        self.set_props(dict(text=text, **kwargs))

    def set_props(self, props):
        self.text = str(props['text'])
        self.size = props.get('size', 'medium')
        self.full_width = props.get('full_width', False)

    def render(self, ctx: dict):
        return {'type': self.type, 'text': self.text, 'size': self.size,
            'fullWidth': self.full_width}



import queue
import typing
import unittest
from collections import deque
from json import dumps, loads

from abstra_internals.controllers.execution import ExecutionController
from abstra_internals.controllers.workflow_engine_detached import DetachedWorkflowEngine
from abstra_internals.repositories.project.project import FormStage
from tests.fixtures import BaseTest


class MockWebSocket:
    msgs: queue.Queue

    def __init__(self, browser_msgs) -> None:
        self.browser_msgs = deque(browser_msgs)
        self.msgs = queue.Queue()

    def send(self, data: str):
        self.msgs.put(["python", loads(data)])

    def receive(self) -> str:
        if len(self.browser_msgs) == 0:
            self.close()

        self.msgs.put(["browser", self.browser_msgs[0]])
        return dumps(self.browser_msgs.popleft())

    def close(self, error_msg: typing.Optional[Exception] = None):
        self.msgs.put(["close", error_msg])

    @property
    def connected(self):
        raise NotImplementedError()


def iter_messages(
    conn: MockWebSocket, msgs: typing.Deque[list], test_case: unittest.TestCase
):
    while True:
        yield
        py_msg = conn.msgs.get(block=True, timeout=2)
        if py_msg[0] == "close":
            if py_msg[1] is None:
                return
            else:
                raise py_msg[1]
        if len(msgs) == 0:
            raise Exception(f"Code is sending unexpected messages: {py_msg}")
        else:
            test_case.maxDiff = None
            test_case.assertEqual(py_msg, msgs[0])
            msgs.popleft()

            if len(msgs) == 0:
                return


def assert_form(
    test_case: BaseTest,
    form_json: FormStage,
    msg_list: list,
    execution_id: str,
):
    msgs: typing.Deque[list] = deque(msg_list)
    browser_msgs = [msg[1] for msg in msgs if msg[0] == "browser"]
    ws = MockWebSocket(browser_msgs)

    ExecutionController(
        repositories=test_case.repositories,
        workflow_engine=DetachedWorkflowEngine(test_case.repositories),
    ).test(
        stage=form_json,
    )

    for msg in iter_messages(ws, msgs, test_case):
        pass



from typing import List, Union

from abstra_internals.entities.forms.page_response import PageResponse


class StepsResponse:
    responses: List[PageResponse]

    @property
    def as_dict(self):
        return self.acc

    def __init__(self):
        self.responses = []
        self.acc = {}

    def __getitem__(self, key: Union[int, str]):
        if isinstance(key, int):
            return self.responses[key]
        if isinstance(key, str):
            return self.acc[key]
        raise IndexError("Step not found")

    def get(self, key: str, default=None):
        return self.acc.get(key, default)

    def append(self, res: PageResponse):
        self.responses.append(res)
        self.__update_acc()

    def __update_acc(self):
        self.acc = {}
        for res in self.responses:
            if isinstance(res, PageResponse):
                self.acc.update(res)
            elif isinstance(res, dict):
                self.acc.update(res)

    def pop(self) -> PageResponse:
        step_response = self.responses.pop()
        self.__update_acc()
        return step_response



from functools import wraps
from threading import Thread


def threaded(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return Thread(name=func.__name__, target=func, args=args, kwargs=kwargs).start()

    return wrapper



import os
import threading
from typing import List

import psutil

from abstra_internals.settings import Settings
from abstra_internals.utils.dot_abstra import RESOURCES_FILE


def get_usage_recursive(pid: int):
    p = psutil.Process(pid)

    cpu = p.cpu_percent()
    mem = int(p.memory_info().rss / 1024 / 1024)

    for child in p.children():
        ch_cpu, ch_mem = get_usage_recursive(child.pid)
        cpu += ch_cpu
        mem += ch_mem

    return cpu, mem


class ResourcesRepository:
    lock = threading.RLock()

    @staticmethod
    def get_file():
        file = Settings.root_path / RESOURCES_FILE
        file.touch(exist_ok=True)
        return file

    @classmethod
    def clear_resources(cls):
        with cls.lock:
            with cls.get_file().open("w") as f:
                f.write("")

    @classmethod
    def save_usage(cls):
        with cls.lock:
            cpu, mem = get_usage_recursive(os.getpid())

            with cls.get_file().open("a") as f:
                f.write(f"{mem} {cpu}\n")

    @classmethod
    def get_all_usage(cls):
        with cls.lock:
            with cls.get_file().open("r") as f:
                mems: List[int] = []
                cpus: List[float] = []

                for line in f.readlines():
                    try:
                        mem, cpu = line.split()
                        mems.append(int(mem))
                        cpus.append(float(cpu))
                    except Exception:
                        pass

                return mems, cpus



import flask

from abstra_internals.controllers.main import MainController
from abstra_internals.controllers.web_editor import WebEditorController


def get_web_editor_bp(main_controller: MainController):
    bp = flask.Blueprint("webEditor", __name__)
    editorAuthController = WebEditorController(main_controller.web_editor_repository)

    @bp.post("/welcome")
    def _welcome():
        data = flask.request.json
        if not data or not data.get("token"):
            flask.abort(400)

        return editorAuthController.welcome(data["token"])

    return bp


def get_editor_api_guard(main_controller: MainController):
    editorAuthController = WebEditorController(main_controller.web_editor_repository)

    return editorAuthController.guard



import pathlib

from abstra_internals.environment import FILES_FOLDER, PROJECT_URL
from abstra_internals.settings import Settings
from abstra_internals.utils.dot_abstra import PERSISTENT_DIR


def get_persistent_dir() -> pathlib.Path:
    """
    Get the path to the persistent directory

    Returns:
        pathlib.Path: Path to the persistent directory
    """

    if FILES_FOLDER:
        return pathlib.Path(FILES_FOLDER)

    dir = Settings.root_path / PERSISTENT_DIR
    dir.mkdir(parents=True, exist_ok=True)
    return dir


def get_public_dir() -> pathlib.Path:
    """
    Get the path to the public directory

    Returns:
        pathlib.Path: Path to the public directory
    """

    return get_persistent_dir() / "_public"


def get_uploads_dir() -> pathlib.Path:
    """
    Get the path to the uploads directory

    Returns:
        pathlib.Path: Path to the uploads directory
    """

    return get_persistent_dir() / "_uploads"


def get_project_url() -> str:
    """
    Get the project URL. localhost if running locally

    Returns:
        str: Project URL
    """

    return PROJECT_URL or f"http://localhost:{Settings.server_port}"


STAGE_RUN_ID_PARAM_KEY = "abstra-run-id"

ABSTRA_LOGO_URL = (
    "https://abstra-cloud-assets.s3.us-east-1.amazonaws.com/logo-small-256px.png"
)



from abstra_internals.controllers.execution_store import ExecutionStore
from abstra_internals.interface.sdk.connectors.google_credentials import (
    create_abstra_google_credentials,
)
from abstra_internals.repositories.connectors import AccessTokenDTO


def get_access_token(name: str) -> AccessTokenDTO:
    sdk_ctx = ExecutionStore.get_by_thread()
    return sdk_ctx.repositories.connectors.get_access_token(name)


def get_gsheets_credentials(connection_name: str = "google-sheets"):
    return create_abstra_google_credentials(connection_name)


def get_gdrive_credentials(connection_name: str = "google-drive"):
    return create_abstra_google_credentials(connection_name)


def get_gmail_credentials(connection_name: str = "gmail"):
    return create_abstra_google_credentials(connection_name)


def get_bigquery_credentials(connection_name: str = "google-bigquery"):
    return create_abstra_google_credentials(connection_name)


def get_google_cal_credentials(connection_name: str = "google-calendar"):
    return create_abstra_google_credentials(connection_name)



import abstra.ai as ai
import abstra.forms as af
import abstra.workflows as aw

### 📌 Abstra Forms is the easiest way to create dynamic forms that allow users to interact with your Python script.

# 💡 Use read functions to receive input types, like multiple choice, text, boolean, and more (check the docs for the full list):
best_movie = af.read_multiple_choice(
    "What is the best movie ever made?",
    options=[
        "Die Hard",
        "Devil Wears Prada",
        "Weekend at Bernie's",
        "American Psycho",
        "Legally Blond",
        "Wall-E",
        "Threat Level: Midnight",
    ],
)

# 💡 Use Python to control logic and the user interactions:
if best_movie == "Threat Level: Midnight":
    af.display("*Michael Scarn theme song plays* Congratulations! You're a genius.")
elif best_movie == "Die Hard":
    af.display("Yep, that's the one. Passed the vibe check.")
else:
    af.display(
        "Nope. The correct answer was 'Die Hard'. Not off to a great start here."
    )

# 💡 Explore more input widgets like read_toggle to get a True/False response:
interview_opt_in = af.read_checkbox(
    "Do you agree to participate in a interview with Prison Mike?", required=False
)

# 💡 Or even read_camera function to prompt the user to submit a photo:
object_photo = af.read_camera(
    "Quick, take a photo of the object on your left. Do it now! 📸"
)

### 📌 Abstra AI is a powerful tool that allows you to prompt an AI model to extract data from images, generate text and more

# 💡 Use the prompt function to send your prompt, attach files, add instructions and specify the response format:
generated_joke = ai.prompt(
    [
        "Write the punchline that Michael Scott from Dunder Mifflin would say if he saw this object on someone's desk in their office. Make sure the joke includes a quirky, slightly deprecating derogatory observation.",
        object_photo,
    ],
    instructions="Only the punchline, and keep it office-friendly!",
)

rating = af.read_rating(f"Rate this joke's funniness: {generated_joke}", char="🤡")

### 📌 Abstra Workflows is a powerful tool that allows you to store and share data between different stages of your workflow

# 💡 Use the set_data function to store data that you want to use later on:
aw.set_data("best_movie", best_movie)
aw.set_data("interview_opt_in", interview_opt_in)
aw.set_data("rating", rating)



import gc
import traceback
from pathlib import Path
from typing import Literal, Optional, Tuple

from abstra_internals.controllers.execution_client import ExecutionClient
from abstra_internals.controllers.execution_client_form import ClientAbandoned
from abstra_internals.controllers.execution_store import ExecutionStore
from abstra_internals.controllers.workflow_interface import IWorkflowEngine
from abstra_internals.entities.execution import Execution
from abstra_internals.logger import AbstraLogger
from abstra_internals.modules import import_as_new
from abstra_internals.repositories.factory import Repositories
from abstra_internals.repositories.project.project import ActionStage
from abstra_internals.usage import execution_usage
from abstra_internals.utils.datetime import now_str

DEFAULT_STATUS = "failed"


def ExecutionTarget(
    *,
    stage: ActionStage,
    execution: Execution,
    client: ExecutionClient,
    repositories: Repositories,
    workflow_engine: IWorkflowEngine,
):
    ExecutionStore.set(execution, client, repositories)
    status = DEFAULT_STATUS

    try:
        repositories.execution.create(execution)
        print(f"[ABSTRA] {now_str()} - Execution started")

        client.handle_start(execution.id)

        status, exception = _execute_code(stage.file_path)
        if exception:
            client.handle_failure(exception)
        else:
            client.handle_success()

    except ClientAbandoned:
        status = "abandoned"
    except Exception as e:
        status = "failed"
        AbstraLogger.capture_exception(e)
    finally:
        print(f"[ABSTRA] {now_str()} - Execution {status}")

        try:
            repositories.stage_run.change_status(execution.stage_run_id, status)
            execution.set_status(status)
            repositories.execution.update(execution)

            workflow_engine.handle_execution_end(execution)
        except Exception as e_final:
            AbstraLogger.capture_exception(e_final)
        finally:
            ExecutionStore.clear()


def _execute_without_exit(filepath: Path):
    try:
        import_as_new(filepath.as_posix())
    except SystemExit as e:
        no_errors = e.code is None or e.code == 0
        if no_errors:
            return
        raise Exception(f"SystemExit: {e.code}") from e


Result = Tuple[Literal["finished", "abandoned", "failed"], Optional[Exception]]


def print_exception(exception: Exception, entrypoint: Path):
    tb = exception.__traceback__
    # tb.tb_frame.f_code.co_filename returns "c:" instead of "C:" on Windows
    while tb and str(entrypoint).lower() != tb.tb_frame.f_code.co_filename.lower():
        tb = tb.tb_next
    traceback.print_exception(type(exception), exception, tb)


@execution_usage
def _execute_code(filepath: Path) -> Result:
    try:
        _execute_without_exit(filepath)
        return "finished", None
    except ClientAbandoned as e:
        return "abandoned", e
    except Exception as e:
        print_exception(e, entrypoint=filepath)
        return "failed", e
    finally:
        gc.collect()



from pathlib import Path

import flask
import flask_sock

from abstra_internals.constants import STAGE_RUN_ID_PARAM_KEY, get_public_dir
from abstra_internals.controllers.execution import (
    ExecutionController,
    LockFailedException,
    UnsetThreadException,
)
from abstra_internals.controllers.execution_client_form import FormClient
from abstra_internals.controllers.execution_client_hook import HookClient
from abstra_internals.controllers.main import MainController
from abstra_internals.entities.execution import context_from_flask
from abstra_internals.environment import (
    BUILD_ID,
    IS_PRODUCTION,
    OIDC_AUTHORITY,
    OIDC_CLIENT_ID,
    SHOW_WATERMARK,
    SIDECAR_SHARED_TOKEN,
)
from abstra_internals.jwt_auth import USER_AUTH_HEADER_KEY
from abstra_internals.logger import AbstraLogger
from abstra_internals.server.cache.control import Cache
from abstra_internals.server.guards.role_guard import (
    Guard,
    PathArgSelector,
    QueryArgSelector,
)
from abstra_internals.server.routes import access_control as ac_router
from abstra_internals.server.routes import auth as auth_router
from abstra_internals.server.routes import kanban as kanban_router
from abstra_internals.server.routes import stage_runs as stage_runs_router
from abstra_internals.server.routes import users as user_router
from abstra_internals.server.routes import workflows as workflows_router
from abstra_internals.server.utils import send_from_dist
from abstra_internals.settings import Settings
from abstra_internals.usage import player_usage
from abstra_internals.utils import check_is_url
from abstra_internals.utils.file import (
    get_random_filepath,
    get_tmp_upload_dir,
    path2module,
)


def get_player_bp(controller: MainController):
    guard = Guard(controller.users_repository, enabled=IS_PRODUCTION)
    cache = Cache(enabled=IS_PRODUCTION)

    bp = flask.Blueprint("player", __name__)
    sock = flask_sock.Sock(bp)

    auth_bp = auth_router.get_player_bp(controller)
    bp.register_blueprint(auth_bp, url_prefix="/_auth")

    user_bp = user_router.get_player_bp(controller)
    bp.register_blueprint(user_bp, url_prefix="/_user")

    kanban_bp = kanban_router.get_player_bp(controller)
    bp.register_blueprint(kanban_bp, url_prefix="/_kanban")

    workflow_bp = workflows_router.get_player_bp(controller)
    bp.register_blueprint(workflow_bp, url_prefix="/_workflows")

    stage_run_bp = stage_runs_router.get_player_bp(controller)
    bp.register_blueprint(stage_run_bp, url_prefix="/_stage-runs")

    access_control_bp = ac_router.get_player_bp(controller)
    bp.register_blueprint(access_control_bp, url_prefix="/_access-control")

    @bp.route("/_healthcheck")
    def _healthcheck():
        return "ok"

    @bp.get("/_workspace")
    def _get_workspace():
        auth = flask.request.headers.get(USER_AUTH_HEADER_KEY)
        return guard.filtered_workspace(auth).as_dict

    @bp.get("/_pages/<string:path>")
    @guard.by(PathArgSelector("path"))
    @player_usage
    def _get_page(path):
        form = controller.get_form_by_path(path)
        if not form:
            flask.abort(404)

        auth = flask.request.headers.get(USER_AUTH_HEADER_KEY)
        return {
            form.type_name: {
                **form.browser_runner_dto,
                "workspace": guard.filtered_workspace(auth).as_dict,
            }
        }

    @bp.get("/_version")
    def _get_version():
        return BUILD_ID

    @bp.get("/_settings")
    def _get_settings():
        return flask.jsonify(
            {
                "show_watermark": SHOW_WATERMARK,
                "oidc_authority": OIDC_AUTHORITY(),
                "oidc_client_id": OIDC_CLIENT_ID(),
            }
        )

    @sock.route("/_socket")
    @guard.socket_by(QueryArgSelector("id"))
    def _websocket(ws: flask_sock.Server):
        request_context = context_from_flask(flask.request)

        try:
            id = flask.request.args.get("id")
            if id is None:
                return

            form = controller.get_form(id)
            if not form:
                return

            client = FormClient(
                request_context=request_context,
                production_mode=True,
                ws=ws,
            )

            ExecutionController(
                repositories=controller.repositories,
                workflow_engine=controller.workflow_engine,
            ).run(
                stage=form,
                client=client,
                request=request_context,
                target_stage_run_id=flask.request.args.get(STAGE_RUN_ID_PARAM_KEY),
            )

        except LockFailedException:
            pass
        except UnsetThreadException:
            pass
        except Exception as e:
            AbstraLogger.capture_exception(e)
        finally:
            ws.close(message="Done")

    @bp.put("/_files")
    @player_usage
    def _upload_file():
        files = flask.request.files
        filename = flask.request.form.get("filename")
        if filename is None:
            flask.abort(400)

        if len(files) == 0:
            flask.abort(400)

        paths = []
        for file in files.values():
            name, path = get_random_filepath(filename)
            file.save(path)
            paths.append(name)
        return paths

    @bp.get("/_files/<path:path>")
    def _get_file(path):
        return flask.send_from_directory(get_tmp_upload_dir(), path)

    @bp.get("/_public/<path:path>")
    def _get_public_file(path):
        return flask.send_from_directory(get_public_dir(), path)

    @bp.get("/_assets/favicon.ico")
    @cache.assets()
    def _favicon():
        favicon_path = controller.get_workspace().favicon_url
        if not favicon_path:
            return _logo()

        if check_is_url(favicon_path):
            return flask.redirect(favicon_path)

        return send_from_dist(favicon_path, dist_folder=Settings.root_path)

    @bp.get("/_assets/logo")
    @cache.assets()
    def _logo():
        logo_path = controller.get_workspace().logo_url
        if not logo_path:
            return flask.abort(404)

        if check_is_url(logo_path):
            return flask.redirect(logo_path)

        return send_from_dist(logo_path, dist_folder=Settings.root_path)

    @bp.get("/_assets/background")
    @cache.assets()
    def _background():
        background_path = controller.get_workspace().theme

        if not background_path:
            return flask.abort(404)

        if check_is_url(background_path):
            return flask.redirect(background_path)

        return send_from_dist(background_path, dist_folder=Settings.root_path)

    @bp.route("/_hooks/<path:path>", methods=["POST", "GET", "PUT", "DELETE", "PATCH"])
    def hook_runner(path):
        hook = controller.get_hook_by_path(path)

        if not hook:
            flask.abort(404)

        if not hook.file:
            flask.abort(500)

        request_context = context_from_flask(flask.request)

        client = HookClient(request_context)

        ExecutionController(
            repositories=controller.repositories,
            workflow_engine=controller.workflow_engine,
        ).run(
            stage=hook,
            client=client,
            request=request_context,
            target_stage_run_id=flask.request.args.get(STAGE_RUN_ID_PARAM_KEY),
        )

        return flask.Response(
            status=client.response["status"],
            response=client.response["body"],
            headers=client.response["headers"],
        )

    @bp.get("/_jobs")
    def list_jobs():
        if flask.request.headers.get("Shared-Token") != SIDECAR_SHARED_TOKEN:
            flask.abort(401)

        jobs = controller.get_jobs()

        # used by sidecar - DO NOT CHANGE CONTRACT
        return [{"id": job.id, "schedule": job.schedule} for job in jobs]

    @bp.post("/_jobs/<path:id>")
    def job_runner(id):
        if flask.request.headers.get("Shared-Token") != SIDECAR_SHARED_TOKEN:
            flask.abort(401)

        job = controller.get_job(id)
        if not job:
            flask.abort(404)

        ExecutionController(
            repositories=controller.repositories,
            workflow_engine=controller.workflow_engine,
        ).submit(stage=job)

        return {"status": "running"}

    @bp.get("/")
    @cache.statics()
    def index():
        res = send_from_dist("player.html", "player.html")
        return res

    @bp.get("/<path:filename>")
    @cache.statics()
    def spa(filename: str):
        res = send_from_dist(filename, "player.html")
        return res

    setup_hook = Path("__setup__.py")
    if setup_hook.exists():
        module = __import__(path2module(setup_hook))
        if hasattr(module, "setup"):
            module.setup(bp)
        else:
            print(f"Could not find setup function in {setup_hook}")

    return bp



from abstra_internals.widgets.widget_base import Output


class MarkdownOutput(Output):
    type = 'markdown-output'

    def __init__(self, text: str, **kwargs):
        self.set_props(dict(text=text, **kwargs))

    def set_props(self, props):
        self.text = props.get('text', '')
        self.full_width = props.get('full_width', False)

    def render(self, ctx: dict):
        return {'type': self.type, 'text': self.text, 'fullWidth': self.
            full_width}



import logging
import threading

from dotenv import load_dotenv as _load_dotenv
from werkzeug.debug import DebuggedApplication
from werkzeug.serving import make_server

from abstra_internals.controllers.execution_consumer import ExecutionConsumer
from abstra_internals.controllers.main import MainController
from abstra_internals.editor_reloader import LocalReloader
from abstra_internals.environment import HOST
from abstra_internals.fs_watcher import watch_file_change_events
from abstra_internals.interface.cli.messages import serve_message
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.consumer import EditorConsumer
from abstra_internals.repositories.factory import get_editor_repositories
from abstra_internals.repositories.producer import LocalProducerRepository
from abstra_internals.resources_watcher import resources_polling_loop
from abstra_internals.server.apps import get_local_app
from abstra_internals.settings import Settings
from abstra_internals.stdio_patcher import StdioPatcher
from abstra_internals.utils.browser import browser_open_editor
from abstra_internals.version import check_latest_version


def start_consumer(controller: MainController):
    if not isinstance(controller.producer_repository, LocalProducerRepository):
        raise ValueError("Invalid producer repository")

    consumer = EditorConsumer(controller.producer_repository.queue)

    th = threading.Thread(
        daemon=True,
        name="execution_consumer",
        target=ExecutionConsumer,
        kwargs=dict(controller=controller, consumer=consumer),
    )

    th.start()

    return consumer, th


def start_file_watcher():
    threading.Thread(
        daemon=True,
        name="file_watcher",
        target=watch_file_change_events,
        kwargs=dict(path=str(Settings.root_path)),
    ).start()


def start_resources_watcher(controller: MainController):
    threading.Thread(
        daemon=True,
        name="resources_watcher",
        target=resources_polling_loop,
        kwargs=dict(controller=controller),
    ).start()


def editor(
    debug: bool,
    load_dotenv: bool,
    reloading: bool,
):
    serve_message()
    check_latest_version()

    controller = MainController(repositories=get_editor_repositories())
    controller.reset_repositories()
    StdioPatcher.apply(controller)
    AbstraLogger.init("local")

    start_consumer(controller)
    start_file_watcher()
    start_resources_watcher(controller)

    if load_dotenv:
        _load_dotenv(Settings.root_path / ".env")

    app = get_local_app(controller)
    if not debug:
        log = logging.getLogger("werkzeug")
        log.setLevel(logging.WARNING)
    else:
        app = DebuggedApplication(app, evalex=True)

    server = make_server(host=HOST, port=Settings.server_port, threaded=True, app=app)

    LocalReloader.set_server(server)

    if not reloading:
        browser_open_editor()

    server.serve_forever()



from unittest import TestCase

from abstra_internals.repositories.project.json_migrations.migration_005 import (
    Migration005,
)


class TestMigration05(TestCase):
    def test_remove_title_from_conditions(self):
        m = Migration005(
            {
                "version": "4.0",
                "conditions": [{"title": "it should not be here anymore"}],
            }
        )
        m.apply()
        self.assertEqual(m.data, {"version": "5.0", "conditions": [{}]})

    def test_remove_title_from_iterators(self):
        m = Migration005(
            {
                "version": "4.0",
                "iterators": [{"title": "it should not be here anymore"}],
            }
        )
        m.apply()
        self.assertEqual(m.data, {"version": "5.0", "iterators": [{}]})

    def test_it_should_not_break_with_no_conditions_and_iterators(self):
        m = Migration005({"version": "4.0"})
        m.apply()
        self.assertEqual(m.data, {"version": "5.0"})



from abstra_internals.widgets.widget_base import Output


class HtmlOutput(Output):
    type = 'html-output'

    def __init__(self, html: str, **kwargs):
        self.set_props(dict(html=html, **kwargs))

    def set_props(self, props):
        self.html = props.get('html', '')
        self.full_width = props.get('full_width', False)

    def render(self, ctx: dict):
        return {'type': self.type, 'html': self.html, 'fullWidth': self.
            full_width}



import os
import sys
import typing
from pathlib import Path

from werkzeug.serving import BaseWSGIServer


class EditorReloader:
    _instance: typing.Optional["EditorReloader"] = None
    _server: typing.Optional[BaseWSGIServer] = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(EditorReloader, cls).__new__(cls)
        return cls._instance

    def set_server(self, server: BaseWSGIServer):
        EditorReloader._server = server

    def reload(self):
        if EditorReloader._server is not None:
            EditorReloader._server.shutdown()
            EditorReloader._server = None

        is_win = sys.platform.startswith("win")
        path_str = sys.argv[0]
        if is_win:
            path_str = path_str + ".exe"

        is_executable_in_args = Path(path_str).exists() and os.access(path_str, os.X_OK)

        executable, cmd = (
            (sys.argv[0], sys.argv[1:])
            if is_executable_in_args
            else (sys.executable, sys.argv)
        )

        cmd = [executable] + cmd
        if "--reloading" not in cmd:
            cmd.append("--reloading")

        if is_win:
            sts = os.spawnv(os.P_WAIT, executable, cmd)
            os._exit(sts)
        else:
            os.execv(executable, cmd)


LocalReloader = EditorReloader()



from unittest import TestCase

from abstra_internals.repositories.project.json_migrations.migration_011 import (
    Migration011,
)


class TestMigration07(TestCase):
    def test_adds_language_to_english(self):
        m = Migration011(
            {
                "version": "10.0",
            }
        )
        m.apply()
        self.assertEqual(
            m.data,
            {
                "version": "11.0",
                "language": "en",
            },
        )



import time

from abstra_internals.controllers.main import MainController
from abstra_internals.services.resources import ResourcesRepository


def resources_polling_loop(*, controller: MainController):
    ResourcesRepository.clear_resources()
    while True:
        try:
            ResourcesRepository.save_usage()
            time.sleep(2)
        except Exception:
            pass



from abstra_internals.linter.linter import LinterFix, LinterIssue, LinterRule
from abstra_internals.repositories.project.project import (
    FormStage,
    Project,
    ProjectRepository,
)
from abstra_internals.utils.file import generate_conflictless_path, is_path_in_conflict


class ConflictingPathFix(LinterFix):
    label = "Fix conflicting path"
    description = "Change the path of the form to avoid conflicts"

    def __init__(self, form: FormStage, project: Project):
        self.form = form
        self.project = project

    def fix(self):
        self.project.update_stage(
            self.form, {"path": generate_conflictless_path(self.form.path)}
        )
        ProjectRepository.save(self.project)


class ConflictingPathFound(LinterIssue):
    type = "bug"
    fixes = []

    def __init__(self, form: FormStage, project: Project):
        self.label = f"The path of the form {form.title} is in conflict with an internal reserved path. This can cause unexpected behavior. You can either change it manually in the Editor or use the 'Fix conflicting path' button."
        self.fixes = [ConflictingPathFix(form, project)]


class ConflictingPath(LinterRule):
    label = "Conflicting path"
    type = "bug"

    def find_issues(self):
        project = ProjectRepository.load()

        issues = []
        for form in project.forms:
            if is_path_in_conflict(form.path):
                issues.append(ConflictingPathFound(form, project))

        return issues



from typing import List
from abstra_internals.widgets.widget_base import AbstraOption, Input, OptionsHandler


class ChecklistInput(Input):
    type = 'checklist-input'
    empty_value: List[str] = []

    def __init__(self, key: str, label: str, options: List[AbstraOption],
        **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, options=options, **kwargs))

    def set_props(self, props):
        self.label = props.get('label', 'Label')
        self.options = props.get('options', [])
        self.value = props.get('initial_value', self.empty_value)
        self.required = props.get('required', True)
        self.hint = props.get('hint', None)
        self.full_width = props.get('full_width', False)
        self.disabled = props.get('disabled', False)
        self.min = props.get('min', 0)
        self.max = props.get('max', len(self.options))
        self.options_handler = OptionsHandler(self.options)

    def is_value_unset(self):
        return False

    def _validate_required(self) ->List[str]:
        if not self.required and len(self.value) == 0:
            return []
        errors = []
        if len(self.value) < self.min:
            if len(self.options) == 1:
                errors.append('i18n_error_required_field')
            else:
                errors.append('i18n_error_min_list')
        if len(self.value) > self.max:
            errors.append('i18n_error_max_list')
        return errors

    def render(self, ctx: dict):
        return {'type': self.type, 'key': self.key, 'options': self.
            options_handler.serialized_options(), 'label': self.label,
            'value': self.serialize_value(), 'required': self.required,
            'hint': self.hint, 'fullWidth': self.full_width, 'disabled':
            self.disabled, 'errors': self.errors, 'min': self.min, 'max':
            self.max}

    def serialize_value(self) ->List:
        return self.options_handler.ids_from_values(self.value)

    def parse_value(self, value):
        return self.options_handler.values_from_ids(value)



import io
import unittest
from dataclasses import dataclass
from datetime import datetime
from typing import List
from uuid import UUID

import requests

from abstra.tables import insert
from abstra_internals.controllers.execution_client_form import FormClient
from abstra_internals.controllers.execution_store import ExecutionStore
from abstra_internals.entities.execution import Execution, RequestContext
from abstra_internals.entities.forms.page_response import PageResponse
from abstra_internals.interface.sdk.forms.step import StepsResponse
from abstra_internals.interface.sdk.tables.api import (
    _make_delete_query,
    _make_insert_multiple_query,
    _make_insert_query,
    _make_row_dict,
    _make_select_query,
    _make_update_query,
    serialize,
)
from abstra_internals.repositories.tables import TablesApiHttpClient
from tests.fixtures import BaseTest


class MockTablesApi(TablesApiHttpClient):
    def __init__(self):
        super().__init__("FAKE_URL")

    def execute(self, query: str, params: List) -> requests.Response:
        resp = requests.Response()
        resp.status_code = 200
        resp.raw = io.BytesIO(
            b'{"errors": null, "returns": {"result": [{"bar": "baz", "car": "cat"}], "fields": [{"name": "bar"}, {"name": "car"}]}}'
        )

        return resp


class TestTables(BaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.repositories.tables = MockTablesApi()
        request_context = RequestContext(
            body="", query_params={}, headers={}, method="GET"
        )
        self.client = FormClient(
            request_context=request_context,
            production_mode=False,
            ws=None,  # type: ignore
        )
        execution = Execution.create(
            request_context=None,
            stage_run_id="mock_sr_id",
            stage_id="mock_stage_id",
        )
        ExecutionStore.set(execution, self.client, self.repositories)

    def tearDown(self) -> None:
        ExecutionStore.clear()
        super().tearDown()

    def test_select(self):
        query, params = _make_select_query(table="foo")
        self.assertEqual(query, 'SELECT * FROM "foo"')
        self.assertEqual(params, [])

    def test_select_with_none(self):
        query, params = _make_select_query(table="foo", where={"bar": None, "baz": 1})
        self.assertEqual(query, 'SELECT * FROM "foo" WHERE "bar" IS NULL AND "baz"=$1')
        self.assertEqual(params, [1])

    def test_select_all_params(self):
        query, params = _make_select_query(
            table="foo", where={"bar": "baz"}, order_by="bar", limit=10, offset=0
        )
        self.assertEqual(
            query,
            'SELECT * FROM "foo" WHERE "bar"=$1 ORDER BY "bar" ASC LIMIT 10 OFFSET 0',
        )
        self.assertEqual(params, ["baz"])

    def test_select_and(self):
        query, params = _make_select_query(
            table="fruits", where={"tangerina": 123456, "melancia": 888}
        )
        self.assertEqual(
            query,
            'SELECT * FROM "fruits" WHERE "tangerina"=$1 AND "melancia"=$2',
        )
        self.assertEqual(params, [123456, 888])

    def test_select_order_desc(self):
        query, params = _make_select_query(
            table="fruits", order_by="tangerina", order_desc=True
        )
        self.assertEqual(
            query,
            'SELECT * FROM "fruits" ORDER BY "tangerina" DESC',
        )
        self.assertEqual(params, [])

    def test_select_crashes_with_non_numeric_limit(self):
        with self.assertRaises(Exception):
            _make_select_query(table="foo", limit="a")  # type: ignore

    def test_select_crashes_with_non_numeric_offset(self):
        with self.assertRaises(Exception):
            _make_select_query(table="foo", offset="a")  # type: ignore

    def test_insert(self):
        query, params = _make_insert_query(table="foo", values={"bar": "baz"})
        self.assertEqual(query, 'INSERT INTO "foo" ("bar") VALUES ($1) RETURNING *')
        self.assertEqual(params, ["baz"])

    def test_insert_datetime(self):
        query, params = _make_insert_query(
            table="foo", values={"bar": datetime(2001, 1, 1, 10, 10, 10, 0)}
        )
        self.assertEqual(query, 'INSERT INTO "foo" ("bar") VALUES ($1) RETURNING *')
        self.assertEqual(params, ["2001-01-01T10:10:10"])

    def test_insert_multiple(self):
        query, params = _make_insert_multiple_query(
            table="foo", values=[{"bar": "baz"}, {"bar": "cat"}]
        )
        self.assertEqual(
            query, 'INSERT INTO "foo" ("bar") VALUES ($1), ($2) RETURNING *'
        )
        self.assertEqual(params, ["baz", "cat"])

    def test_insert_multiple_with_empty_values(self):
        query, params = _make_insert_multiple_query(table="foo", values=[])
        self.assertEqual(query, 'INSERT INTO "foo" DEFAULT VALUES RETURNING *')
        self.assertEqual(params, [])

    def test_insert_multiple_with_missing_values(self):
        query, params = _make_insert_multiple_query(
            table="foo", values=[{"bar": "baz", "foo": 1}, {"foo": 2}]
        )
        self.assertEqual(
            query,
            'INSERT INTO "foo" ("bar", "foo") VALUES ($1, $2), (DEFAULT, $3) RETURNING *',
        )
        self.assertEqual(params, ["baz", 1, 2])

    def test_insert_multiple_with_missing_and_nulls(self):
        query, params = _make_insert_multiple_query(
            table="foo", values=[{"bar": None, "foo": 1}, {"bar": "baz"}]
        )
        self.assertEqual(
            query,
            'INSERT INTO "foo" ("bar", "foo") VALUES ($1, $2), ($3, DEFAULT) RETURNING *',
        )
        self.assertEqual(params, [None, 1, "baz"])

    def test_insert_dict(self):
        query, params = _make_insert_query(
            table="foo", values={"bar": {"bar": 1, "foo": "expected"}}
        )
        self.assertEqual(query, 'INSERT INTO "foo" ("bar") VALUES ($1) RETURNING *')
        self.assertEqual(params, ["""{"bar": 1, "foo": "expected"}"""])

    def test_insert_list(self):
        query, params = _make_insert_query(table="foo", values={"bar": [1, 2, 3, 4]})
        self.assertEqual(query, 'INSERT INTO "foo" ("bar") VALUES ($1) RETURNING *')
        self.assertEqual(params, ["""[1, 2, 3, 4]"""])

    def test_insert_set(self):
        query, params = _make_insert_query(
            table="foo", values={"bar": set([1, 2, 3, 4])}
        )
        self.assertEqual(query, 'INSERT INTO "foo" ("bar") VALUES ($1) RETURNING *')
        self.assertEqual(params, ["""[1, 2, 3, 4]"""])

    def test_insert_tuples(self):
        query, params = _make_insert_query(table="foo", values={"bar": (1, 2, 3, 4)})
        self.assertEqual(query, 'INSERT INTO "foo" ("bar") VALUES ($1) RETURNING *')
        self.assertEqual(params, ["""[1, 2, 3, 4]"""])

    def test_insert_default_values(self):
        query, params = _make_insert_query(table="foo", values={})
        self.assertEqual(query, 'INSERT INTO "foo" DEFAULT VALUES RETURNING *')
        self.assertEqual(params, [])

    def test_insert_returning(self):
        row = insert(table="foo", values={"bar": "baz", "car": "cat"})
        self.assertEqual(row, {"bar": "baz", "car": "cat"})

    def test_update(self):
        query, params = _make_update_query(
            table="foo", set={"bar": "baz"}, where={"id": 1}
        )
        self.assertEqual(query, 'UPDATE "foo" SET "bar"=$1 WHERE "id"=$2 RETURNING *')
        self.assertEqual(params, ["baz", 1])
        self.assertEqual(params, ["baz", 1])

    def test_update_with_none_in_where(self):
        query, params = _make_update_query(
            table="foo", set={"a": "b"}, where={"c": None, "d": 1}
        )
        self.assertEqual(
            query, 'UPDATE "foo" SET "a"=$1 WHERE "c" IS NULL AND "d"=$2 RETURNING *'
        )
        self.assertEqual(params, ["b", 1])

    def test_update_with_multiple_where_conditions(self):
        query, params = _make_update_query(
            table="foo", set={"bar": "baz"}, where={"id": 1, "name": "foo"}
        )
        self.assertEqual(
            query, 'UPDATE "foo" SET "bar"=$1 WHERE "id"=$2 AND "name"=$3 RETURNING *'
        )
        self.assertEqual(params, ["baz", 1, "foo"])

    def test_update_two_items(self):
        query, params = _make_update_query(
            table="foo", set={"bar": "baz", "car": "cat"}, where={"id": 1}
        )
        self.assertEqual(
            query, 'UPDATE "foo" SET "bar"=$1, "car"=$2 WHERE "id"=$3 RETURNING *'
        )
        self.assertEqual(params, ["baz", "cat", 1])

    def test_update_without_where(self):
        with self.assertRaises(Exception):
            _make_update_query(table="foo", set={"bar": "baz"}, where={})

    def test_update_without_set(self):
        with self.assertRaises(Exception):
            _make_update_query(table="foo", set={}, where={"id": 1})

    def test_delete(self):
        query, params = _make_delete_query(table="foo", values={"id": 1})
        self.assertEqual(query, 'DELETE FROM "foo" WHERE "id"=$1 RETURNING *')
        self.assertEqual(params, [1])

    def test_delete_with_none(self):
        query, params = _make_delete_query(table="foo", values={"bar": None, "baz": 1})
        self.assertEqual(
            query, 'DELETE FROM "foo" WHERE "bar" IS NULL AND "baz"=$1 RETURNING *'
        )
        self.assertEqual(params, [1])

    def test_delete_and(self):
        query, params = _make_delete_query(
            table="fruits", values={"tangerina": 123456, "melancia": 888}
        )
        self.assertEqual(
            query,
            'DELETE FROM "fruits" WHERE "tangerina"=$1 AND "melancia"=$2 RETURNING *',
        )
        self.assertEqual(params, [123456, 888])

    def test_escaping_table_name(self):
        query, params = _make_delete_query(
            table='table "with" quotes', values={'"tangerina"': "123456"}
        )
        self.assertEqual(
            query,
            'DELETE FROM "table ""with"" quotes" WHERE """tangerina"""=$1 RETURNING *',
        )

    # security tests
    def test_delete_sql_injection(self):
        query, params1 = _make_delete_query(
            table="fruits", values={"tangerina": "12345678; DROP TABLE fruits"}
        )
        query2, params2 = _make_delete_query(
            table="fruits", values={"tangerina": "123456"}
        )
        self.assertEqual(query, query2)

    def test_delete_without_values(self):
        with self.assertRaises(Exception):
            _make_delete_query(table="foo", values={})


class TestMakeRowDict(unittest.TestCase):
    def test_dict(self):
        row = _make_row_dict({"bar": "baz", "car": "cat"})
        self.assertEqual(row, {"bar": "baz", "car": "cat"})

    def test_dataclass(self):
        @dataclass
        class Foo:
            bar: str
            car: str

        row = _make_row_dict(Foo(bar="baz", car="cat"))
        self.assertEqual(row, {"bar": "baz", "car": "cat"})

    def test_page_response(self):
        row = _make_row_dict(PageResponse({"bar": "baz", "car": "cat"}, None))
        self.assertEqual(row, {"bar": "baz", "car": "cat"})

    def test_step_response(self):
        pr1 = PageResponse({"a": "1", "b": 2}, None)
        pr2 = PageResponse({"b": "3", "c": 4}, None)
        steps_response = StepsResponse()
        steps_response.append(pr1)
        steps_response.append(pr2)
        row = _make_row_dict(steps_response)
        self.assertEqual(row, {"a": "1", "b": "3", "c": 4})

    def test_none(self):
        row = _make_row_dict(None)
        self.assertEqual(row, {})


class TestSerialize(unittest.TestCase):
    def test_datetime(self):
        value = datetime(2001, 1, 1, 10, 10, 10, 0)
        self.assertEqual(serialize(value), "2001-01-01T10:10:10")

    def test_set(self):
        value = set([1, 2, 3, 4])
        self.assertEqual(serialize(value), "[1, 2, 3, 4]")

    def test_list(self):
        value = [1, 2, 3, 4]
        self.assertEqual(serialize(value), "[1, 2, 3, 4]")

    def test_dict(self):
        value = {"bar": 1, "foo": "expected"}
        self.assertEqual(serialize(value), """{"bar": 1, "foo": "expected"}""")

    def test_tuple(self):
        value = (1, 2, 3, 4)
        self.assertEqual(serialize(value), "[1, 2, 3, 4]")

    def test_uuid(self):
        value = UUID("12345678-1234-5678-1234-567812345678")
        self.assertEqual(serialize(value), "12345678-1234-5678-1234-567812345678")



import os
from typing import List, Optional

from dotenv import load_dotenv

from abstra_internals.linter.linter import LinterFix, LinterIssue, LinterRule
from abstra_internals.services.env_vars import EnvVarsRepository
from abstra_internals.settings import Settings


class LoadEnvVars(LinterFix):
    label = "Load environment variables"
    rule: "EnvFileChanged"

    def __init__(self, rule: "EnvFileChanged") -> None:
        self.rule = rule

    def fix(self):
        load_dotenv(Settings.root_path / ".env", override=True)
        self.rule.has_env_file_initially = True


class EnvFileChangeFound(LinterIssue):
    def __init__(self, rule: "EnvFileChanged") -> None:
        self.label = "Your .env file was updated. Please load the new environment variables to ensure consistency."
        self.fixes = [LoadEnvVars(rule)]


class EnvFileChanged(LinterRule):
    label = "Changes to .env file detected"
    type = "bug"
    has_env_file_initially: Optional[bool] = None

    def find_issues(self) -> List[LinterIssue]:
        env_file = Settings.root_path / ".env"
        has_env_file = env_file.exists()

        if self.has_env_file_initially is None:
            self.has_env_file_initially = has_env_file
            if has_env_file:
                return []

        if self.has_env_file_initially != has_env_file:
            if has_env_file:
                # File was created
                return [EnvFileChangeFound(self)]
            else:
                # File was deleted
                # No need to report this behavior as the missing_env rule will take care of it
                self.has_env_file_initially = has_env_file
                return []

        env_vars = EnvVarsRepository.list()
        for env in env_vars:
            if env.value != os.getenv(env.name):
                return [EnvFileChangeFound(self)]

        return []
